function [signals]=signal_gathering(sig_chirp,t,f,CRmax)
% [signals]=signal_gathering(chirpsf,t,f,CRmax);
% gathers individual chirps into continuous wave piecewise linear signals.
%
% Input parameters:
%   
%   sig_chirps      lists significant detected chirps created in
%                   chirpgathering.m in a matrix. Each row
%                   identifies a detected chirp and consists of 4 values:
%
%                   sig_chirps(indchirp,:) = ...
%                      [start time, initial frequency, end time, chirprate] 
%
%   t               the time values for the time indices of the chirplet
%                   transform that generated the sig_chirps list  
%   f               the freq values for the freq indices of the chirplet
%                   transform that generated the sig_chirps list 
%   CRmax           Chirp Rate Maximum
%
% Output variables:
%
%   signals         lists all detected chirps in a matrix. Each row
%                   identifies a detected chirp and consists of 5 values:
%
%                   chirps(indchirp,:) = ...
%                      [start time, initial frequency, end time, chirprate, signal number]
%
%                   where the signal number indicates the FMCW signal to
%                   which the chirp is associated. This runs from 1,...k.
%                   Chirps with signal number 0 are considered unassigned.
%
% Code written by Fabien Millioz and Mike Davies, last update 2016. This
% function is part of a software implementation of the Maximum Chirplet
% Transform and its application to the detection of piecewise linear chirp
% signals. Theory and details of the technique can be found in:
%
%   Millioz and Davies, "Sparse detection in the chirplet transform:
%   application to FMCW radar signals," IEEE Transactions on Sig. Proc. 
%   vol 60(6), pp 2800  2813, 2012.
%
% Please reference this paper in any publication that makes use of this
% software. Copyright 2016, University of Edinburgh.


% define chirp tolerance as in gathering.m
dt=t(2)-t(1);
deltat=3*dt;          % to extend the chirp

l=length(sig_chirp);
timebegs=sig_chirp(:,1)*ones(1,l); % list of chirp time starts
freqbegs=sig_chirp(:,2)*ones(1,l); % list of chirp frequency starts
timeends=ones(l,1)*sig_chirp(:,3)'; % list of chirp time ends
freqends=ones(l,1)* (sig_chirp(:,2)+sig_chirp(:,4).*(sig_chirp(:,3)-sig_chirp(:,1)))'; % list of chirp frequency ends

% normalization
timebegs=timebegs/max(t);
timeends=timeends/max(t);
freqbegs=freqbegs/max(f);
freqends=freqends/max(f);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Check start and end of signals (note we should allow some leaway here -
% such as the number of chirps starting/ending within deltat of the
% start/end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% number of starting chirps
startchirps=sum(sig_chirp(:,1)<=deltat+1);
indstart=find(sig_chirp(:,1)<=deltat+1);
disp(['Number of signals detected at the start of the data: ',num2str(startchirps)]);

% number of finishing chirps
finchirps=sum(sig_chirp(:,3)>=max(sig_chirp(:,3))-deltat);
indfin=find(sig_chirp(:,3)>=max(sig_chirp(:,3))-deltat);
disp(['Number of signals detected at the end of the data: ',num2str(finchirps)]);

% should be the same...
if startchirps~=finchirps
    disp('WARNING: different number of signal at the beginning and at the end');
end


% Calculate pairwise distances between chirp ends and other chirp starts
dist=sqrt( (timeends-timebegs).^2 + (freqends-freqbegs).^2); % pairwise distances

% avoid self assignment - artificially increase diagonal distances
dist=dist+max(max(dist)).*eye(size(dist));

% Define TF proximity measure removing the ending chirps 
TF_proximity=dist(:,find(sig_chirp(:,3)<max(sig_chirp(:,3))-deltat));
TF_proximity(1:startchirps,:)=max(max(TF_proximity))+1; % mask the beginning chirps

% associating minimum distance to each chirp end and identify links
[minval,links]=min(TF_proximity);

% remove links that are not close in TF (accounting for TF uncertainty see
% eqn (52) [Millioz and Davies])
threshold=dt*sqrt(max(t)^-2+(CRmax)^2); 
links(minval>threshold)=0;
linkslength=length(links);

% Check for multiple links to the same chirp - keep closest
for c_indx = 1:l
    if sum(links==c_indx)>1
        duplicate_list = find(links==c_indx);
        [min_v, min_indx] = min(minval(duplicate_list)); % use T proximity measure
        links(duplicate_list) = 0; % unassign links
        links(duplicate_list(min_indx)) = c_indx; % reassign closest link
    end
end     

%Define Time proximity measure
T_proximity=abs(timeends-timebegs);
T_proximity = T_proximity+ +max(max(T_proximity)).*eye(size(T_proximity)); % avoid self assignment
T_proximity([1:startchirps links(links~=0)],:)=max(max(T_proximity)); % mask assigned chirps and start chirps
[minval2,links2]=min(T_proximity);

% fill in unassigned links 
links(links==0)=links2(links==0); 

% Final check for duplicate links to the same chirp - keep closest
for c_indx = 1:l
    if sum(links==c_indx)>1
        duplicate_list = find(links==c_indx);
        [min_v, min_indx] = min(minval2(duplicate_list)); % this time use T proximity measure
        links(duplicate_list) = 0; % unassign links
        links(duplicate_list(min_indx)) = c_indx; % reassign closest link
    end
end   

% assign the signals to the chirps
signals=zeros(l,5); signals(:,1:4)=sig_chirp;

% generate a link list for all sig_chirps (including end chirps)
link_list=zeros(1,length(sig_chirp));
link_list(sig_chirp(:,3)<max(sig_chirp(:,3))-deltat)=links;

for ind = 1:startchirps
    curind = indstart(ind);            
    signals(curind,5)=ind; 
    while sum(indfin==curind)==0 
        if link_list(curind)==0
            disp('WARNING: premature termination of a signal (it did not reach the end of the data)');
            break;
        end
        curind=link_list(curind);
        signals(curind,5)=ind; 
    end    
end

% wrap up
disp([num2str(sum(signals(:,5)~=0)),' chirps assigned to signals']);
disp([num2str(sum(signals(:,5)==0)),' chirps unassigned']);

